#!/usr/bin/env python3

import sys
import json

def creature_validate(data):
    errors = []
    if isinstance(data, dict):
        for key, value in data.items():
            if key == 'name':
                if isinstance(value, dict):
                    for key2, value2 in value.items():
                        if not key2 in ['singular', 'plural']:
                            errors.append('unknown key "' + key2 + '" in name')
                        elif not isinstance(value2, str):
                            errors.append('name ' + key2 + ' is not a string')
                else:
                    errors.append('name is not an object')
            elif key == 'level': # do levels <= 0 work?
                if not isinstance(value, int):
                    errors.append('level is not an integer')
            elif key == 'faction':
                if not isinstance(value, str):
                    errors.append('faction is not a string')
            elif key == 'cost':
                if isinstance(value, int):
                    if value <= 0:
                        errors.append('cost is not an integer > 0')
                elif isinstance(value, dict):
                    for key2, value2 in value.items():
                        if not isinstance(value2, int) and value2 < 0:
                            errors.append('cost ' + key2 + ' is not an integer >= 0')
                else:
                    errors.append('cost is not an object or integer')
            elif key == 'fightValue': # does 0 work?
                if not isinstance(value, int) or value < 0:
                    errors.append('fightValue is not an integer >= 0')
            elif key == 'aiValue': # does 0 work?
                if not isinstance(value, int) or value < 0:
                    errors.append('aiValue is not an integer >= 0')
            elif key == 'growth':
                if not isinstance(value, int) or value <= 0:
                    errors.append('growth is not an integer > 0')
            elif key == 'hitPoints':
                if not isinstance(value, int) or value <= 0:
                    errors.append('hitPoints is not an integer > 0')
            elif key == 'speed': # does 0 work?
                if not isinstance(value, int) or value < 0:
                    errors.append('speed is not an integer >= 0')
            elif key == 'attack':
                if not isinstance(value, int) or value < 0:
                    errors.append('attack is not an integer >= 0')
            elif key == 'defence':
                if not isinstance(value, int) or value < 0:
                    errors.append('defence is not an integer >= 0')
            elif key == 'damage': # does 0 work?
                if isinstance(value, int):
                    if value < 0:
                        errors.append('damage is not an integer >= 0')
                elif isinstance(value, list) and len(value) == 2 and isinstance(value[0], int) and isinstance(value[1], int):
                    if value[0] <= 0 or value[1] <= 0 or value[0] > value[1]:
                        errors.append('max damage is not >= min damage >= 0')
                else:
                    errors.append('damage is not an integer >= 0 or array of two integers >= 0')
            elif key == 'amm':
                if isinstance(value, int):
                    if not value > 0:
                        errors.append('amm is not an integer > 0')
                elif isinstance(value, list) and len(value) == 2 and isinstance(value[0], int) and isinstance(value[1], int):
                    if value[0] < 0 or value[1] < 0 or value[0] > value[1]:
                        errors.append('max amm is not >= min amm >= 0')
                else:
                    errors.append('amm is not an integer >= 0 or array of two integers >= 0')
            elif key == 'upgrades':
                if isinstance(value, list):
                    for value2 in value:
                        if not isinstance(value2, str):
                            errors.append('upgrades is not an array of strings')
                else:
                    errors.append('upgrades is not an array of strings')
            elif key == 'doubleWide':
                if not isinstance(value, bool):
                    errors.append('doubleWide is not a boolean')
            elif key == 'shots':
                if not isinstance(value, int) or value < 0:
                    errors.append('shots is not an integer >= 0')
            elif key == 'spellPoints':
                if not isinstance(value, int) or value < 0:
                    errors.append('spellPoints is not >= 0')
            elif key == 'graphics':
                if isinstance(value, dict):
                    for key2, value2 in value.items():
                        if key2 == 'animation':
                            if not isinstance(value2, str):
                                errors.append('animation is not a string')
                        elif key2 == 'timeBetweenFidgets': # does 0 work?
                            if not isinstance(value2, (int, float)) or value2 < 0:
                                errors.append('timeBetweenFidgets is not a float >= 0')
                        elif key2 == 'troopCountLocationOffset':
                            if not isinstance(value2, int):
                                errors.append('troopCountLocationOffset is not an integer')
                        elif key2 == 'troopCountLocationOffset':
                            if not isinstance(value2, int) or value2 < 0:
                                errors.append('troopCountLocationOffset is not an integer >= 0')
                        elif key2 == 'animationTime': # does 0 work?
                            if isinstance(value2, dict):
                                for key3, value3 in value2.items():
                                    if not key3 in ['walk', 'attack', 'flight']:
                                        errors.append('unknown key "' + key3 + '" in animationTime')
                                    elif not isinstance(value3, (int, float)) or value3 < 0:
                                        errors.append('animationTime ' + key3 + ' is not a float >= 0')
                            else:
                                errors.append('animationTime is not an object')
                        elif key2 == 'background':
                            if isinstance(value2, dict):
                                for key3, value3 in value2.items():
                                    if not key3 in ['120', '130']:
                                        errors.append('unknown key "' + key3 + '" in background')
                                    elif not isinstance(value3, str):
                                        errors.append('background ' + key3 + ' is not a string')
                            else:
                                errors.append('background is not an object')
                        elif key2 == 'missile':
                            if isinstance(value2, dict):
                                for key3, value3 in value2.items():
                                    if key3 == 'animation':
                                        if not isinstance(value3, str):
                                            errors.append('missile animation is not a string')
                                    elif key3 == 'spin':
                                        if not isinstance(value3, bool):
                                            errors.append('missile spin is not a boolean')
                                    elif key3 == 'offset':
                                        if isinstance(value3, dict):
                                            for key4, value4 in value3.items():
                                                if not key4 in ['upperX', 'upperY', 'middleX', 'middleY', 'lowerX', 'lowerY']:
                                                    errors.append('unknown key "' + key4 + '" in missile offset')
                                                elif not isinstance(value4, int):
                                                    errors.append('missile offset ' + key4 + ' is not an integer')
                                        else:
                                            errors.append('missile offset is not an object')
                                    elif key3 == 'missileFrameAngles':
                                        if isinstance(value3, list): # do there have to be exactly 12?
                                            for value4 in value3:
                                                if not isinstance(value4, (int, float)):
                                                    errors.append('missile frameAngles is not an array of floats')
                                        else:
                                            errors.append('missile frameAngles is not an array of floats')
                                    else:
                                        errors.append('unknown key "' + key3 + '" in missile')
                            else:
                                errors.append('missile is not an object')
                        else:
                            errors.append('unknown key "' + key2 + '" in graphics')
                else:
                    errors.append('graphics is not an object')
            elif key == 'sound':
                if isinstance(value, dict):
                    for key2, value2 in value.items():
                        if not key2 in ['attack', 'defend', 'killed', 'move', 'shoot', 'wince', 'moveStart', 'moveEnd']:
                            errors.append('unknown key "' + key2 + '" in sound')
                        elif not isinstance(value2, str):
                            errors.append('sound ' + key2 + ' is not a string')
                else:
                    errors.append('sound is not an object')
            elif key == 'abilities':
                pass
            else:
                errors.append('unknown key "' + key + '"')
    else:
        errors.append('creature is not an object')
    return errors

def creature_empty():
    return {'upgrades' : [],
            'doubleWide' : False,
            'shots' : 0,
            'spellPoints' : 0,
            'cost' : dict(),
            'abilities' : [],
            'graphics' : {'timeBetweenFidgets' : 1.0,
                          'troopCountLocationOffset' : 0,
                          'attackClimaxFrame' : 0,
                          'animationTime' : {'walk' : 1.0,
                                             'attack' : 1.0,
                                             'flight' : 1.0},
                          'background' : {'120' : '',
                                           '130' : ''},
                          'missile' : {'animation' : '',
                                       'spin' : False,
                                       'offset' : {'upperX' : 0,
                                                   'upperY' : 0,
                                                   'middleX' : 0,
                                                   'middleY' : 0,
                                                   'lowerX' : 0,
                                                   'lowerY' : 0},
                                       'frameAngles' : []}},
            'sound' : {'attack' : '',
                       'defend' : '',
                       'killed' : '',
                       'move' : '',
                       'shoot' : '',
                       'wince' : '',
                       'moveStart' : '',
                       'moveEnd' : ''}}

def creature_merge(old, new):
    if 'name' in new:
        if not 'name' in old:
            old['name'] = new['name']
        else:
            for key, value in new['name'].items():
                old['name'][key] = value
    for key in ['level', 'faction', 'fightValue', 'aiValue', 'growth', 'hitPoints', 'speed', 'attack', 'defence', 'doubleWide', 'shots', 'spellPoints']:
        if key in new:
            old[key] = new[key]
    for key in ['damage', 'amm']:
        if key in new:
            if isinstance(new[key], int):
                new[key] = [new[key], new[key]]
            old[key] = new[key]
    if 'cost' in new:
        if isinstance(new['cost'], int):
            new['cost'] = {'Gold' : new['cost']}
        for key, value in new['cost'].items():
            old['cost'][key] = value
    if 'graphics' in new:
        for key in ['animation', 'timeBetweenFidgets', 'troopCountLocationOffset', 'attackClimaxFrame']:
            if key in new['graphics']:
                old['graphics'][key] = new['graphics'][key]
        if 'animationTime' in new['graphics']:
            for key, value in new['graphics']['animationTime'].items():
                old['graphics']['animationTime'][key] = value
        if 'background' in new['graphics']:
            for key, value in new['graphics']['background'].items():
                old['graphics']['background'][key] = value
        if 'missile' in new['graphics']:
            for key in ['animation', 'spin', 'frameAngles']:
                if key in new['graphics']['missile']:
                    old['graphics']['missile'][key] = new['graphics']['missile'][key]
            if 'offset' in new['graphics']['missile']:
                for key in ['upperX', 'upperY', 'middleX', 'middleY', 'lowerX', 'lowerY']:
                    if key in new['graphics']['missile']['offset']:
                        old['graphics']['missile']['offset'][key] = new['graphics']['missile']['offset'][key]
    if 'sound' in new:
        for key, value in new['sound'].items():
            old['sound'][key] = value

    # need to think about these two
    if 'upgrades' in new:
        old['upgrades'] = new['upgrades']
    if 'abilities' in new:
        old['abilities'] = new['abilities']

def creature_find_all(data):
    errors = []
    if 'name' in data:
        if not 'singular' in data['name']:
            errors.append('singular name missing')
        if not 'plural' in data['name']:
            errors.append('plural name missing')
    else:
        errors.append('name section missing')
    if not 'level' in data:
        errors.append('level missing')
    if 'faction' in data:
        if not data['faction'] in tmpTowns:
            errors.append('faction "' + data['faction'] + '" undefined')
    else:
        errors.append('faction missing')
    if 'cost' in data:
        for resource, amount in data['cost'].items():
            if not resource in tmpResources:
                errors.append('resource "' + resource + '" undefined')
    else:
        errors.append('cost section missing')
    if not 'fightValue' in data:
        errors.append('fightValue missing')
    if not 'aiValue' in data:
        errors.append('aiValue missing')
    if not 'growth' in data:
        errors.append('growth missing')
    if not 'hitPoints' in data:
        errors.append('hitPoints missing')
    if not 'speed' in data:
        errors.append('speed missing')
    if not 'attack' in data:
        errors.append('attack missing')
    if not 'defence' in data:
        errors.append('defence missing')
    if not 'damage' in data:
        errors.append('damage missing')
    if not 'amm' in data:
        errors.append('amm missing')
    if not 'animation' in data['graphics']:
        errors.append('animation missing')
    if 'upgrades' in data:
        for upgrade in data['upgrades']:
            if not upgrade in tmpCreatures:
                errors.append('creature "' + upgrade + '" undefined')
    return errors

modToLoad = sys.argv[1]
mod = json.load(open(modToLoad + '/mod.json'))

mods = mod['mods']

tmpCreatures = dict()
tmpTowns = dict()
tmpResources = {'Gold' : dict()}

for mod in mods:
    config = json.load(open(mod + '/mod.json'))
    if not 'config' in config:
        continue
    config = config['config']
    errors = False

    if 'creature' in config:
        for creatureFile in config['creature']:
            creatures = json.load(open(mod + '/config/' + creatureFile))
            for creature, data in creatures.items():
                e = creature_validate(data)
                if len(e) > 0:
                    print('Mod: ' + mod + ', File: ' + creatureFile + ', Creature: ' + creature)
                    print('\n'.join([ '\t' + v for v in e ]))
                    errors = True
                if not creature in tmpCreatures:
                    tmpCreatures[creature] = creature_empty()
                creature_merge(tmpCreatures[creature], data)

    if 'town' in config:
        for townFile in config['town']:
            towns = json.load(open(mod + '/config/' + townFile))
            for town, data in towns.items():
                tmpTowns[town] = dict()

    if 'resource' in config:
        for resourceFile in config['resource']:
            resources = json.load(open(mod + '/config/' + resourceFile))
            for resource, data in resources.items():
                tmpResources[resource] = dict()

    for creature, data in tmpCreatures.items():
        e = creature_find_all(data)
        if len(e) > 0:
            print('Mod: ' + mod + ', Creature: ' + creature)
            print('\n'.join([ '\t' + v for v in e ]))
            errors = True
    if errors:
        pass
        #exit(1)

print(json.dumps(tmpTowns, indent='\t', separators=(', ', ' : ')))
print(json.dumps(tmpResources, indent='\t', separators=(', ', ' : ')))
print(json.dumps(tmpCreatures, indent='\t', separators=(', ', ' : ')))
